/*

Copyright (C) 2018 Werner Taelemans, Belgium

This file is part of Galjoen.

Galjoen is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; version 3 of the License.

Galjoen is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Galjoen; see the file COPYING.  If not, see
<http://www.gnu.org/licenses/>.

*/

#include"magicmoves.h"
#include<iostream>
#include<iomanip>
#include<algorithm>
#include<chrono>
#include<stdexcept>
using std::logic_error;
using std::cout;
using std::endl;
using std::hex;
using std::dec;




auto possible_occupancies=[](const BitBoard& mask)->vector<BitBoard>
{
  vector<BitBoard> r;
  vector<Square> vs=mask.get_squares();
  BitBoard bb;
  int max=1<<vs.size();
  for(int i=0;i<max;i++){
    bb.clear();
    int state_number=i;
    auto j=vs.begin();
    while(state_number){
      if(state_number&0x01){bb.set_square(*j);}
      state_number>>=1;
      j++;
    }
    r.push_back(bb);
  }
  return r;
};

auto rook_moves=[](const Square& s,const BitBoard& occupancy)->BitBoard
{
  BitBoard temp_bb,r;
  temp_bb.set_square(s);
  for(int i=1;i<8;i++){
    temp_bb=temp_bb<<8;
    r|=temp_bb;
    if(temp_bb&occupancy){break;}
  }
  
  temp_bb.clear();
  temp_bb.set_square(s);
  for(int i=1;i<8;i++){
    temp_bb=temp_bb<<1 & not_a_file;
    r|=temp_bb;
    if(temp_bb&occupancy){break;}
  }
  
  temp_bb.clear();
  temp_bb.set_square(s);
  for(int i=1;i<8;i++){
    temp_bb=temp_bb>>1 & not_h_file;
    r|=temp_bb;
    if(temp_bb&occupancy){break;}
  }
  
  temp_bb.clear();
  temp_bb.set_square(s);
  for(int i=1;i<8;i++){
    temp_bb=temp_bb>>8;
    r|=temp_bb;
    if(temp_bb&occupancy){break;}
  }
  return r;
};

auto bishop_moves=[](const Square& s,const BitBoard& occupancy)->BitBoard
{
  BitBoard temp_bb,r;
  temp_bb.set_square(s);
  for(int i=1;i<8;i++){
    temp_bb=temp_bb<<7 & not_h_file;
    r|=temp_bb;
    if(temp_bb&occupancy){break;}
  }
  
  temp_bb.clear();
  temp_bb.set_square(s);
  for(int i=1;i<8;i++){
    temp_bb=temp_bb<<9 & not_a_file;
    r|=temp_bb;
    if(temp_bb&occupancy){break;}
  }
  
  temp_bb.clear();
  temp_bb.set_square(s);
  for(int i=1;i<8;i++){
    temp_bb=temp_bb>>9 & not_h_file;
    r|=temp_bb;
    if(temp_bb&occupancy){break;}
  }
  
  temp_bb.clear();
  temp_bb.set_square(s);
  for(int i=1;i<8;i++){
    temp_bb=temp_bb>>7 & not_a_file;
    r|=temp_bb;
    if(temp_bb&occupancy){break;}
  }
  return r;
};

auto construct_magic_bishop_moves=[]()->const array<vector<BitBoard>,64>
{
  array<vector<BitBoard>,64> r;
  BishopMask bm;
  MagicNumber mn;
  for(Square s=a1;s<=h8;s++){
    unsigned long long magic_number=mn.get_magic_number<'B'>(s).first;
    int shift_value=mn.get_magic_number<'B'>(s).second;
    r[s.id()].resize(1<<(64-shift_value));
    auto vpo=possible_occupancies(bm.get_mask(s));
    for(auto ppo=vpo.begin();ppo!=vpo.end();++ppo)
      {
	BitBoard mov=bishop_moves(s,*ppo);
	unsigned long long occupancy_number=BitBoard::to_ull(*ppo);
	unsigned int index=occupancy_number*magic_number>>shift_value;
	r[s.id()][index]=mov;
      }
  }
  return r;
};

auto construct_magic_rook_moves=[]()->const array<vector<BitBoard>,64>
{
  array<vector<BitBoard>,64> r;
  RookMask rm;
  MagicNumber mn;
  for(Square s=a1;s<=h8;s++){
    unsigned long long magic_number=mn.get_magic_number<'R'>(s).first;
    int shift_value=mn.get_magic_number<'R'>(s).second;
    r[s.id()].resize(1<<(64-shift_value));
    auto vpo=possible_occupancies(rm.get_mask(s));
    for(auto ppo=vpo.begin();ppo!=vpo.end();++ppo)
      {
	BitBoard mov=rook_moves(s,*ppo);
	unsigned long long occupancy_number=BitBoard::to_ull(*ppo);
	unsigned int index=occupancy_number*magic_number>>shift_value;
	r[s.id()][index]=mov;
      }
  }
  return r;
};

MagicMoves::MagicMoves()
{  
  magic_bishop_moves=construct_magic_bishop_moves();
  magic_rook_moves=construct_magic_rook_moves();

  vector<BitBoard> temp_bishop_attacks;
  for(int sq=0;sq<64;sq++){
    temp_bishop_attacks=magic_bishop_moves[sq];
    sort(temp_bishop_attacks.begin(),temp_bishop_attacks.end());
    bishop_attacks[sq].clear();
    unique_copy(temp_bishop_attacks.begin(),temp_bishop_attacks.end(),back_inserter(bishop_attacks[sq]));
  }
  vector<BitBoard> temp_rook_attacks;
  for(int sq=0;sq<64;sq++){
    temp_rook_attacks=magic_rook_moves[sq];
    sort(temp_rook_attacks.begin(),temp_rook_attacks.end());
    rook_attacks[sq].clear();
    unique_copy(temp_rook_attacks.begin(),temp_rook_attacks.end(),back_inserter(rook_attacks[sq]));
  }

  for(int sq=0;sq<64;sq++){
    for(size_t i=0;i<magic_bishop_moves[sq].size();i++){
      BitBoard bb=magic_bishop_moves[sq][i];
      for(size_t j=0;j<bishop_attacks[sq].size();j++){
	if(bishop_attacks[sq][j]==bb){bishop_index[sq].push_back(static_cast<unsigned char>(j));break;}
      }
    }
    if(bishop_index[sq].size()!=magic_bishop_moves[sq].size()){throw logic_error("MagicMoves error(1) in constructor");}
  }

  for(int sq=0;sq<64;sq++){
    for(size_t i=0;i<magic_rook_moves[sq].size();i++){
      BitBoard bb=magic_rook_moves[sq][i];
      for(size_t j=0;j<rook_attacks[sq].size();j++){
	if(rook_attacks[sq][j]==bb){rook_index[sq].push_back(static_cast<unsigned char>(j));break;}
      }
    }
    if(rook_index[sq].size()!=magic_rook_moves[sq].size()){throw logic_error("MagicMoves error(2) in constructor");}
  }

  //if(!selftest_ok()){throw logic_error("MagicMoves error in selftest");}

  for(int sq=0;sq<64;sq++){
    magic_bishop_moves[sq].clear();magic_bishop_moves[sq].resize(10);
    magic_rook_moves[sq].clear();magic_rook_moves[sq].resize(10);
  }

}

MagicMoves::~MagicMoves()
{  
}

int MagicMoves::size()const
{
  int r{0};
  for(int i=0;i<64;i++){
    r+=magic_bishop_moves[i].size()*sizeof(BitBoard);
    r+=magic_rook_moves[i].size()*sizeof(BitBoard);
    r+=bishop_attacks[i].size()*sizeof(BitBoard);
    r+=rook_attacks[i].size()*sizeof(BitBoard);
    r+=bishop_index[i].size();
    r+=rook_index[i].size();    
  }
  return r;        // size in bytes
}

bool MagicMoves::selftest_ok()const
{
  bool r{true};

  //-------------------------- values in attack tables unique? --------------------------
  //
  for(int sq=0;sq<64;sq++){
    for(size_t i=0;i<bishop_attacks[sq].size();i++){
      for(size_t j=i+1;j<bishop_attacks[sq].size();j++){
	if(bishop_attacks[sq][i]==bishop_attacks[sq][j]){
	  cout << endl << "MagicMoves::selftest_ok() error in bishopattacks"
	       << "   square=" << sq << "   i=" << i << "   j=" << j << endl;
	  return false;
	}
      }
    }
    for(size_t i=0;i<rook_attacks[sq].size();i++){
      for(size_t j=i+1;j<rook_attacks[sq].size();j++){
	if(rook_attacks[sq][i]==rook_attacks[sq][j]){
	  cout << endl << "MagicMoves::selftest_ok() error in rookattacks"
	       << "   square=" << sq << "   i=" << i << "   j=" << j << endl;
	  return false;
	}
      }
    }
  }
  //-------------------------------------------------------------------------------------
  
  
  for(size_t sq=0;sq<64;sq++){
    for(size_t i=0;i<bishop_index[sq].size();i++){
      if(bishop_attacks[sq][bishop_index[sq][i]]!=magic_bishop_moves[sq][i]){
	cout << "square=" << sq << endl;
	cout << "index=" << i << endl;
	cout << "magic_bishop_moves: " << hex << BitBoard::to_ull(magic_bishop_moves[sq][i]) << dec << endl;
	cout << "bishop_attacks: " << hex << BitBoard::to_ull(bishop_attacks[sq][bishop_index[sq][i]]) << dec << endl;
	return false;
      }
    }
  }
  for(size_t sq=0;sq<64;sq++){
    for(size_t i=0;i<rook_index[sq].size();i++){
      if(rook_attacks[sq][rook_index[sq][i]]!=magic_rook_moves[sq][i]){
	cout << "square=" << sq << endl;
	cout << "index=" << i << endl;
	cout << "magic_rook_moves: " << hex << BitBoard::to_ull(magic_rook_moves[sq][i]) << dec << endl;
	cout << "rook_attacks: " << hex << BitBoard::to_ull(rook_attacks[sq][rook_index[sq][i]]) << dec << endl;
	return false;
      }
    }
  }

  return r;
}
