/*

Copyright (C) 2018 Werner Taelemans, Belgium

This file is part of Galjoen.

Galjoen is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; version 3 of the License.

Galjoen is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Galjoen; see the file COPYING.  If not, see
<http://www.gnu.org/licenses/>.

*/

#include"../board/cposition.h"
#include"../eval/mobility.h"
#include"../eval/evaluation.h"
#include"../common/numerals.h"
#include"../common/prebb.h"
#include<cmath>
#include<iostream>
#include<sstream>
#include<string>
#include<stdexcept>
#include<cassert>
using std::round;
using std::logic_error;
using std::cout;
using std::endl;
using std::string;
using std::ostringstream;



Evaluation::Evaluation(const ChessPosition& x,Egt4Pieces& y1,Egt4Pieces& y2): cp(x), m_egtkbnk(y1), m_egtkqkr(y2), mob(Mobility(x))
{
  m_variable=vector<float>(m_weight.size(),0.0);
  m_pst.init(x);
}

Evaluation::~Evaluation()
{
}

void Evaluation::init()
{
  m_pst.init(cp);
}

void Evaluation::clear()
{
  m_egtkbnk.clear();
}

void Evaluation::set_log(bool x)
{
  m_log=x;
}

string Evaluation::fen()const
{
  return cp.fen();
}

void Evaluation::make_move(const Move& m)
{
  m_pst.make_move(m);
}

void Evaluation::unmake_move(const Move& m)
{
  m_pst.unmake_move(m);
}

int Evaluation::static_eval()
{
  m_eval_count++;

  m_score=0;
  compute_material();
  compute_mobility();
  compute_pawn_eval();
  compute_development();
  compute_pst();
  compute_king_safety();
  compute_king_activity();
  compute_pinned_pieces();
  compute_opposite_colored_bishops();
  m_variable[w_stm] = (cp.stm()*2-1);

  float middlegame=cp.middlegame();
  float endgame=1.0-middlegame;
  for(unsigned w=w_material;w<=w_stm;w++){
    m_score+=m_weight.mg(w)*m_variable[w]*middlegame + m_weight.eg(w)*m_variable[w]*endgame;
  }
  m_score/=(1.0+m_variable[w_opposite_colored_bishops]);
  m_score=round(m_score);

  
  //------------------------------------------------------------
  // Special endgames:
  //
  switch(cp.materialindex().index()){
    //white
  case 0x0011000000000:m_score=KBNK();return m_score*(2*cp.stm()-1);
  case 0x0020000000000:m_score=KBBK();return m_score*(2*cp.stm()-1);
  case 0x0002000000000:m_score=KNNK();return m_score*(2*cp.stm()-1);
  case 0x0010100000000:m_score=KBPK();return m_score*(2*cp.stm()-1);
  case 0x1000000001000:m_score=KQKR();return m_score*(2*cp.stm()-1);
  case 0x1000000000000:m_score=KQK();return m_score*(2*cp.stm()-1);
  case 0x0100000000000:m_score=KRK();return m_score*(2*cp.stm()-1);
  case 0x0000100000000:m_score=KPK();return m_score*(2*cp.stm()-1);
  case 0x0100100001000:m_score=KRPKR();return m_score*(2*cp.stm()-1);
    //black
  case 0x0000000000110:m_score=KBNK();return m_score*(2*cp.stm()-1);
  case 0x0000000000200:m_score=KBBK();return m_score*(2*cp.stm()-1);
  case 0x0000000000020:m_score=KNNK();return m_score*(2*cp.stm()-1);
  case 0x0000000000101:m_score=KBPK();return m_score*(2*cp.stm()-1);
  case 0x0100000010000:m_score=KQKR();return m_score*(2*cp.stm()-1);
  case 0x0000000010000:m_score=KQK();return m_score*(2*cp.stm()-1);
  case 0x0000000001000:m_score=KRK();return m_score*(2*cp.stm()-1);
  case 0x0000000000001:m_score=KPK();return m_score*(2*cp.stm()-1);
  case 0x0100000001001:m_score=KRPKR();return m_score*(2*cp.stm()-1);
  }
  
  //
  //------------------------------------------------------------

  return m_score * m_variable[w_stm];
}


//----------------- compute raw eval terms: ----------------------
//                  ++++++++++++++++++++++
//
//
void Evaluation::compute_material()
{
  MaterialIndex mi=cp.materialindex();
  //std::cerr << "Evaluation::compute_material()   m_mater.get_eval(" << mi.str() << ")=" << m_mater.get_eval(mi) << "   m_weight[0]=" << m_weight.eg(0) << endl;
  m_variable[w_material]=m_mater.get_eval(mi);
}

void Evaluation::compute_mobility()
{
  // general safe mobility:
  mob.compute();
  m_variable[w_mobility]=mob.safe_mobility();

  // penalty for locked up rook:
  m_variable[w_mobility]+=10*(((cp.black_rook() & g8h8h7) && (cp.black_king() & f8g8))+((cp.black_rook() & a8b8c8a7) && (cp.black_king() & b8c8d8))
	       -((cp.white_rook() & g1h1h2) && (cp.white_king() & f1g1))-((cp.white_rook() & a1b1c1a2) && (cp.white_king() & b1c1d1)));

  // undefended_pieces:
  m_variable[w_undefended_pieces]=mob.undefended_pieces();

  // hanging pieces:
  m_variable[w_hanging_pieces]=mob.hanging_pieces();
}

void Evaluation::compute_development()
{
  m_variable[w_development]=0;

  // minor pieces:
  m_variable[w_development]-=( (cp.white_knight() | cp.white_bishop()) & rank1).popcount()*20;
  m_variable[w_development]+=( (cp.black_knight() | cp.black_bishop()) & rank8).popcount()*20;
  
  // rooks:
  m_variable[w_development]+= (cp.white_rook() & (bb_white_open_files)).popcount()*20;
  m_variable[w_development]-= (cp.black_rook() & (bb_black_open_files)).popcount()*20;

  // early queen development:
  if(cp.white_queen() & rank3_8){m_variable[w_development]-=10;}
  if(cp.black_queen() & rank6_1){m_variable[w_development]+=10;}

  // blocked centre pawns:
  if((d2.bitboard() & cp.white_pawn()) && (d3.bitboard() & cp.pieces())){m_variable[w_development]-=20;} 
  if((e2.bitboard() & cp.white_pawn()) && (e3.bitboard() & cp.pieces())){m_variable[w_development]-=20;} 
  if((d7.bitboard() & cp.black_pawn()) && (d6.bitboard() & cp.pieces())){m_variable[w_development]+=20;} 
  if((e7.bitboard() & cp.black_pawn()) && (e6.bitboard() & cp.pieces())){m_variable[w_development]+=20;}

}

void Evaluation::compute_pst()
{
  m_variable[w_pst_pawn_mg]=m_pst.get_pawn_mg()*cp.middlegame();
  m_variable[w_pst_pawn_eg]=m_pst.get_pawn_eg()*cp.endgame();
  m_variable[w_pst_knight_mg]=m_pst.get_knight_mg()*cp.middlegame();
  m_variable[w_pst_knight_eg]=m_pst.get_knight_eg()*cp.endgame();
  m_variable[w_pst_bishop_mg]=m_pst.get_bishop_mg()*cp.middlegame();
  m_variable[w_pst_bishop_eg]=m_pst.get_bishop_eg()*cp.endgame();
  m_variable[w_pst_rook_mg]=m_pst.get_rook_mg()*cp.middlegame();
  m_variable[w_pst_rook_eg]=m_pst.get_rook_eg()*cp.endgame();
  m_variable[w_pst_queen_mg]=m_pst.get_queen_mg()*cp.middlegame();
  m_variable[w_pst_queen_eg]=m_pst.get_queen_eg()*cp.endgame();
  m_variable[w_pst_king_mg]=m_pst.get_king_mg()*cp.middlegame();
  m_variable[w_pst_king_eg]=m_pst.get_king_eg()*cp.endgame();
}

void Evaluation::compute_king_safety()
{
  // castling
  int castling_rights=20*static_cast<bool>(cp.castlingrights() & 0xff);
  castling_rights-=20*static_cast<bool>(cp.castlingrights()>>8 & 0xff);

  // pawn shield
  m_white_pawn_shield=((white_king_neighbourhood() | white_king_neighbourhood()<<8)&cp.white_pawn()).popcount();
  m_black_pawn_shield=((black_king_neighbourhood() | black_king_neighbourhood()>>8)&cp.black_pawn()).popcount();
  int pawn_shield=10*(m_white_pawn_shield - m_black_pawn_shield);

  // pawn storm
  int pawn_storm=5*(black_king_battlefield() & cp.white_pawn()).popcount();
  pawn_storm-=5*(white_king_battlefield() & cp.black_pawn()).popcount();

  // open file on king
  int open_files_on_king=15*(black_king_neighbourhood() & bb_white_open_files).popcount();
  open_files_on_king-=15*(white_king_neighbourhood() & bb_black_open_files).popcount();

  
  // attacks on the king:
  m_variable[w_king_safety]=mob.king_safety()/256.0;

  //
  m_variable[w_king_safety]+=castling_rights;
  m_variable[w_king_safety]+=pawn_shield;
  m_variable[w_king_safety]+=pawn_storm;
  m_variable[w_king_safety]+=open_files_on_king;
  if(cp.check()){
    m_variable[w_king_safety]-=pawn_value*(2*cp.stm()-1);
  }
  m_variable[w_king_safety]*=cp.middlegame();
}

void Evaluation::compute_king_activity()
{
  int wka{0},bka{0};
  if(!(cp.white_pawn() | cp.black_pawn())){
    wka=static_cast<bool>(cp.white_king()&centre)+static_cast<bool>(cp.white_king()&bigcentre)-static_cast<bool>(cp.white_king()&border);
    bka=static_cast<bool>(cp.black_king()&centre)+static_cast<bool>(cp.black_king()&bigcentre)-static_cast<bool>(cp.black_king()&border);    
  }else{
    wka-=distance_between(cp.white_king().get_first_square(),bb_black_weak_pawns)*2;
    bka-=distance_between(cp.black_king().get_first_square(),bb_white_weak_pawns)*2;
    //
    wka-=distance_between(cp.white_king().get_first_square(),bb_white_passed_pawn_telestoppers)*3;
    bka-=distance_between(cp.black_king().get_first_square(),bb_black_passed_pawn_telestoppers)*3;
  }
  m_variable[w_king_activity]=(wka-bka)*cp.endgame();
}

void Evaluation::compute_pawn_eval()
{
  uint32_t ph=cp.get_pawnhash();
  if(m_phtr.get_key()!=ph){
    m_new_pawn_eval++;
    if(m_pht.read(ph)==false){
      m_missed_pht_read++;
      m_phtr.set_key(ph);
      m_phtr.set_pawns(cp.white_pawn(),cp.black_pawn());
      m_pht.set_current(m_phtr);
      m_pht.write();
    }
  
    m_phtr=m_pht.get_current();
    bb_white_passed_pawns=m_phtr.get_white_passed_pawns();
    bb_white_passed_pawn_stoppers=m_phtr.get_white_passed_pawn_stoppers();
    bb_white_passed_pawn_telestoppers=m_phtr.get_white_passed_pawn_telestoppers();
    bb_white_weak_pawns=m_phtr.get_white_weak_pawns();
    bb_white_open_files=m_phtr.get_white_open_files();
    bb_black_passed_pawns=m_phtr.get_black_passed_pawns();
    bb_black_passed_pawn_stoppers=m_phtr.get_black_passed_pawn_stoppers();
    bb_black_passed_pawn_telestoppers=m_phtr.get_black_passed_pawn_telestoppers();
    bb_black_weak_pawns=m_phtr.get_black_weak_pawns();
    bb_black_open_files=m_phtr.get_black_open_files();
  
    m_white_passed_pawns=m_phtr.get_weighted_white_passed_pawns();
    m_black_passed_pawns=m_phtr.get_weighted_black_passed_pawns();
    m_variable[w_doubled_pawns]=m_phtr.get_weighted_doubled_pawns();
    m_variable[w_pawn_islands]=m_phtr.get_pawn_islands();
    m_variable[w_rookpawn]=m_phtr.get_rook_pawns();
    m_variable[w_weak_pawns]=m_phtr.get_weak_pawns();
    m_variable[w_isolated_pawns]=m_phtr.get_isolated_pawns();
  }

  //-------------------------------------- blocked pawns: --------------------------------------
  //
  m_variable[w_blocked_pawns]=(cp.black_pawn()>>8 & cp.pieces()).popcount() - (cp.white_pawn()<<8 & cp.pieces()).popcount();
  

  //-------------------------------------- bad bishop: -----------------------------------------
  //
  m_variable[w_bad_bishop]=0;
  if(cp.black_bishop().popcount()==1){
    if(cp.black_bishop() & white_squares){
      m_variable[w_bad_bishop]+=(cp.black_pawn() & white_squares).popcount();      
    }else{
      m_variable[w_bad_bishop]+=(cp.black_pawn() & black_squares).popcount();      
    }
  }
  if(cp.white_bishop().popcount()==1){
    if(cp.white_bishop() & white_squares){
      m_variable[w_bad_bishop]-=(cp.white_pawn() & white_squares).popcount();      
    }else{
      m_variable[w_bad_bishop]-=(cp.white_pawn() & black_squares).popcount();      
    }
  }
  

  //-------------------------------------- passed pawns: ---------------------------------------
  //                         
  int rw=m_white_passed_pawns;
  if(rw!=0){
    if(distance_between(cp.black_king().get_first_square(),bb_white_passed_pawns)<2){rw-=150;}
    else if(distance_between(cp.black_king().get_first_square(),bb_white_passed_pawns)<3){rw-=75;}
    else if(distance_between(cp.black_king().get_first_square(),bb_white_passed_pawns)<4){rw-=35;}
    else if(distance_between(cp.black_king().get_first_square(),bb_white_passed_pawns)<5){rw-=15;}
    rw-=(bb_white_passed_pawn_stoppers & cp.black_pieces() & ~cp.black_rook()).popcount()*50;
    BitBoard wpps{bb_white_passed_pawn_stoppers};
    while(wpps){
      if(wpps.get_reset_square() & mob.black_attacks() & ~mob.white_attacks()){
	rw-=50;
      }
    }

    //----------------------------- black king in square of white passed pawns? ------------------
    if(!(cp.black_queen() | cp.black_rook() | cp.black_bishop() | cp.black_knight())){
      BitBoard wpp=(bb_white_passed_pawns & rank7)<<8;  // put the passed pawns on their promotion square
      while(wpp){
	if(distance_between(cp.black_king().get_first_square(),wpp.get_reset_square())>1){rw+=pawn_value*4;}
      }
      wpp=(bb_white_passed_pawns & rank6)<<16;
      while(wpp){
	if(distance_between(cp.black_king().get_first_square(),wpp.get_reset_square())>2){rw+=pawn_value*3;}
      }
      wpp=(bb_white_passed_pawns & rank5)<<24;
      while(wpp){
	if(distance_between(cp.black_king().get_first_square(),wpp.get_reset_square())>3){rw+=pawn_value*2;}
      }
      wpp=(bb_white_passed_pawns & rank4)<<32;
      while(wpp){
	if(distance_between(cp.black_king().get_first_square(),wpp.get_reset_square())>4){rw+=pawn_value;}
      }
      wpp=(bb_white_passed_pawns & rank3)<<40;
      while(wpp){
	if(distance_between(cp.black_king().get_first_square(),wpp.get_reset_square())>5){rw+=pawn_value/2;}
      }
    }
    if(rw<0)rw=0;
  }

  int rb=m_black_passed_pawns;
  if(rb!=0){
    if(distance_between(cp.white_king().get_first_square(),bb_black_passed_pawns)<2){rb-=150;}
    else if(distance_between(cp.white_king().get_first_square(),bb_black_passed_pawns)<3){rb-=75;}
    else if(distance_between(cp.white_king().get_first_square(),bb_black_passed_pawns)<4){rb-=35;}
    else if(distance_between(cp.white_king().get_first_square(),bb_black_passed_pawns)<5){rb-=15;}
    rb-=(bb_black_passed_pawn_stoppers & cp.white_pieces() & ~cp.white_rook()).popcount()*50;
    BitBoard bpps{bb_black_passed_pawn_stoppers};
    while(bpps){
      if(bpps.get_reset_square() & mob.white_attacks() & ~mob.black_attacks()){
	rb-=50;
      }
    }

    //----------------------------- white king in square of black passed pawns? ------------------
    if(!(cp.white_queen() | cp.white_rook() | cp.white_bishop() | cp.white_knight())){
      BitBoard bpp=(bb_black_passed_pawns & rank2)>>8;  // put the passed pawns on their promotion square
      while(bpp){
	if(distance_between(cp.white_king().get_first_square(),bpp.get_reset_square())>1){rb+=pawn_value*4;}
      }
      bpp=(bb_black_passed_pawns & rank3)>>16;
      while(bpp){
	if(distance_between(cp.white_king().get_first_square(),bpp.get_reset_square())>2){rb+=pawn_value*3;}
      }
      bpp=(bb_black_passed_pawns & rank4)>>24;
      while(bpp){
	if(distance_between(cp.white_king().get_first_square(),bpp.get_reset_square())>3){rb+=pawn_value*2;}
      }
      bpp=(bb_black_passed_pawns & rank5)>>32;
      while(bpp){
	if(distance_between(cp.white_king().get_first_square(),bpp.get_reset_square())>4){rb+=pawn_value;}
      }
      bpp=(bb_black_passed_pawns & rank6)>>40;
      while(bpp){
	if(distance_between(cp.white_king().get_first_square(),bpp.get_reset_square())>5){rb+=pawn_value/2;}
      }
    }
    if(rb<0)rb=0;
  }
  
  int r=rw-rb;
  m_variable[w_passed_pawns]=r;
  //--------------------------------------------------------------------------------------------

}

void Evaluation::compute_pinned_pieces()
{
  m_variable[w_pinned_pieces]=(cp.black_pieces() & cp.pinned_pieces()).popcount() - (cp.white_pieces() & cp.pinned_pieces()).popcount();
}

void Evaluation::compute_opposite_colored_bishops()
{
  m_variable[w_opposite_colored_bishops]=0.0;
  BitBoard bb{cp.white_bishop() | cp.black_bishop()};
  if(bb.popcount()==2 && cp.white_bishop().popcount()==1){
    if(  ((bb & white_squares) != bb) && ((bb & black_squares) != bb)  ){
      m_variable[w_opposite_colored_bishops]=cp.endgame();
    }
  }
}




//----------------- interface to raw eval terms: ----------------------
//                  +++++++++++++++++++++++++++
//
//
int Evaluation::material()const
{
  return m_variable[w_material];
}

int Evaluation::mobility()
{
  return m_variable[w_mobility];
}

int Evaluation::king_safety()const
{
  return m_variable[w_king_safety];
}

BitBoard Evaluation::passed_pawns()const
{
  return bb_white_passed_pawns | bb_black_passed_pawns;
}



//----------------- interface to weighted eval terms: ----------------------
//                  +++++++++++++++++++++++++++++++++
//
//

vector<float>& Evaluation::variable_vector()
{
  return m_variable;
}

vector<float>& Evaluation::weight_mg_vector()
{
  return m_weight.weight_mg_vector();
}

vector<float>& Evaluation::weight_eg_vector()
{
  return m_weight.weight_eg_vector();
}

//------------------------------------------------------------------------------------

int Evaluation::distance_between(Square x,Square y)const
{
  return m_d(x,y);
}

int Evaluation::distance_between(Square s,BitBoard b)const
{
  int dist=8;
  BitBoard bsquares=b;
  while(bsquares)
    {
      Square bs=bsquares.get_reset_square().get_unique_square();
      int x=distance_between(s,bs);
      if(x<dist){dist=x;}
    }
  return dist;
}

const PieceSquareTables& Evaluation::pst()const
{
  return m_pst;
}

string Evaluation::stats()
{
  ostringstream oss;
  oss << "Evaluation::stats)  missed pht reads: " << m_missed_pht_read << "   total evals: " << m_eval_count << endl;
  return oss.str();
}

string Evaluation::report()
{
  int eval=static_eval();
  if(!cp.white_to_move()){eval=-eval;}
  ostringstream ostr;
  ostr << endl
       << "Static eval (white):       " << setw(10) << eval << endl
       << "Material:                  " << setw(10) << m_weight.mg(w_material)*m_variable[w_material] << "       " << setw(10) << m_weight.eg(w_material)*m_variable[w_material] << endl
       << "Mobility:                  " << setw(10) << m_weight.mg(w_mobility)*m_variable[w_mobility] << "       " << setw(10) << m_weight.eg(w_mobility)*m_variable[w_mobility] << endl
       << "Development:               " << setw(10) << m_weight.mg(w_development)*m_variable[w_development] << "       " << setw(10) << m_weight.eg(w_development)*m_variable[w_development] << endl
       << "Piece Square Tables:       " << setw(10) << m_weight.mg(w_pst_knight_mg)*m_variable[w_pst_knight_mg] << "       " << setw(10) << m_weight.eg(w_pst_knight_mg)*m_variable[w_pst_knight_mg] << endl
       << "King safety:               " << setw(10) << m_weight.mg(w_king_safety)*m_variable[w_king_safety] << "       " << setw(10) << m_weight.eg(w_king_safety)*m_variable[w_king_safety] << endl
       << "King activity:             " << setw(10) << m_weight.mg(w_king_activity)*m_variable[w_king_activity] << "       " << setw(10) << m_weight.eg(w_king_activity)*m_variable[w_king_activity] << endl
       << "Rook pawn:                 " << setw(10) << m_weight.mg(w_rookpawn)* m_variable[w_rookpawn] << "       " << setw(10) << m_weight.eg(w_rookpawn)* m_variable[w_rookpawn] << endl
       << "Doubled pawns:             " << setw(10) << m_weight.mg(w_doubled_pawns)*m_variable[w_doubled_pawns] << "       " << setw(10) << m_weight.eg(w_doubled_pawns)*m_variable[w_doubled_pawns] << endl
       << "Weak pawns:                " << setw(10) << m_weight.mg(w_weak_pawns)*m_variable[w_weak_pawns] << "       " << setw(10) << m_weight.eg(w_weak_pawns)*m_variable[w_weak_pawns] << endl
       << "Isolated pawns:            " << setw(10) << m_weight.mg(w_isolated_pawns)*m_variable[w_isolated_pawns] << "       " << setw(10) << m_weight.eg(w_isolated_pawns)*m_variable[w_isolated_pawns] << endl
       << "Passed pawns:              " << setw(10) << m_weight.mg(w_passed_pawns)*m_variable[w_passed_pawns] << "       " << setw(10) << m_weight.eg(w_passed_pawns)*m_variable[w_passed_pawns] << endl
       << "Pawn islands:              " << setw(10) << m_weight.mg(w_pawn_islands)*m_variable[w_pawn_islands] << "       " << setw(10) << m_weight.eg(w_pawn_islands)*m_variable[w_pawn_islands] << endl
       << "Pinned pieces:             " << setw(10) << m_weight.mg(w_pinned_pieces)*m_variable[w_pinned_pieces] << "       " << setw(10) << m_weight.eg(w_pinned_pieces)*m_variable[w_pinned_pieces] << endl
       << "Opposite colored bishops:  " << setw(10) << m_weight.mg(w_opposite_colored_bishops)*m_variable[w_opposite_colored_bishops] << "       "
       << setw(10) << m_weight.eg(w_opposite_colored_bishops)*m_variable[w_opposite_colored_bishops] << endl
       << "Side to move bonus:        " << setw(10) << m_weight.mg(w_stm)*(cp.stm()*2-1) << "       " << setw(10) << m_weight.eg(w_stm)*(cp.stm()*2-1) << endl;
  return ostr.str();
}

//------------------------ Special endgames ---------------------------
//
int Evaluation::KQK()
{
  string s=cp.fen();
  int r=m_egtkqkr.eval(s);
  if(r!=0xFF){return r;}
  r=abs(m_score);

  int sign{1};
  BitBoard losingking, winningking;
  if(cp.black_queen()){
    sign=-1;
    losingking=cp.white_king();
    winningking=cp.black_king();
  }else{
    losingking=cp.black_king();
    winningking=cp.white_king();
  }

  if(losingking & border){r+=1000;}
  else if(losingking & border_minus1){r+=900;}
  else if(losingking & border_minus2){r+=800;}

  r-=10*distance_between(losingking.get_unique_square(),winningking.get_unique_square());

  return r*sign;
}

int Evaluation::KRK()
{
  string s=cp.fen();
  int r=m_egtkqkr.eval(s);
  if(r!=0xFF){return r;}
  r=abs(m_score);
  
  int sign{1};
  BitBoard losingking, winningking;
  if(cp.black_rook()){
    sign=-1;
    losingking=cp.white_king();
    winningking=cp.black_king();
  }else{
    losingking=cp.black_king();
    winningking=cp.white_king();
  }

  if(losingking & border){r+=1000;}
  else if(losingking & border_minus1){r+=900;}
  else if(losingking & border_minus2){r+=800;}

  r-=10*distance_between(losingking.get_unique_square(),winningking.get_unique_square());
  
  return r*sign;
}

int Evaluation::KNNK()const
{
  int r{0};
  BitBoard lk=cp.black_knight()?cp.white_king():cp.black_king();
  if(lk&corner){r=100;}

  if(cp.black_knight()){r=-r;}
  return r;
}

int Evaluation::KBBK()const
{
  int r{1000};
  Square wk,lk;  // winning king, loosing king

  if(cp.black_bishop()){
    if((cp.black_bishop()&white_squares).popcount()==2 ||(cp.black_bishop()&black_squares).popcount()==2){return 0;}
    lk=cp.white_king().get_first_square();
    wk=cp.black_king().get_first_square();
  }else{
    if((cp.white_bishop()&white_squares).popcount()==2 ||(cp.white_bishop()&black_squares).popcount()==2){return 0;}
    lk=cp.black_king().get_first_square();
    wk=cp.white_king().get_first_square();
  }

  r-=distance_between(wk,bigcentre|c2f2b3g3b6g6c7f7)*100;
  r-=distance_between(lk,corner)*10;
  r+=distance_between(lk,centre)*10;
  r-=distance_between(lk,wk);
  
  if(cp.black_bishop()){r=-r;}
  return r;
}

int Evaluation::KBNK()
{
  string s=cp.fen();
  int r=m_egtkbnk.eval(s);
  if(r!=0xFF){return r;}
  r=600;

  int sign{1};
  BitBoard losingking, winningking,winningknight,winningbishop;
  if(cp.black_knight()){
    sign=-1;
    losingking=cp.white_king();
    winningking=cp.black_king();
    winningknight=cp.black_knight();
    winningbishop=cp.black_bishop();
  }else{
    losingking=cp.black_king();
    winningking=cp.white_king();
    winningknight=cp.white_knight();
    winningbishop=cp.white_bishop();
  }

  if(losingking & border){r+=60;}
  else if(losingking & border_minus1){r+=40;}
  else if(losingking & border_minus2){r+=20;}

  if(winningbishop & white_squares){
    r-=10*distance_between(losingking.get_unique_square(),white_corner);
  }else{
    r-=10*distance_between(losingking.get_unique_square(),black_corner);
  }

  r-=distance_between(losingking.get_unique_square(),winningking.get_unique_square());  
  return r*sign;
}

int Evaluation::KPK()const
{
  int r;
  int rank{3},dtp;
  bool b_flip{false};
  bool wtm=cp.white_to_move();
  BitBoard K,k,P;
  if(cp.black_pawn()){
    K=cp.black_king();
    k=cp.white_king();
    P=cp.black_pawn();
    K.flip();k.flip();P.flip();wtm=!wtm;b_flip=true;
  }else{
    K=cp.white_king();
    k=cp.black_king();
    P=cp.white_pawn();
  }
  if(P & kings_half){
    K.flop();k.flop();P.flop();
  }

  if(P & (rank2 | rank3)){rank=3;}  
  if(P & rank4){rank=4;}  
  if(P & rank5){rank=5;}  
  if(P & rank6){rank=6;}  
  if(P & rank7){rank=7;}  
  dtp=8-rank;
  r=rank*100;

  
  if(!(P&a_file)){
    if(distance_between(k.get_first_square(),P)<=dtp &&
       distance_between(k.get_first_square(),P)<distance_between(K.get_first_square(),P) &&
       !wtm){r=0;}
    else if(rank<7 && ((P<<8 & k) || (P<<16 & k))){r=0;}
    else if(rank<5 && (P<<8 & K) && (P<<24 & k) && wtm){r=0;}
    else if(rank==7 && ((P<<7 & k) || P<<9 & k)){r=0;}
    else {r+=distance_between(k.get_first_square(),P);}
  }else{
    r=distance_between(k.get_first_square(),a8);
    if(r==1){r=0;}
  }
  
  if(b_flip){r=-r;}
  return r;
}

int Evaluation::KBPK()const
{
  int r;
  bool wtm=cp.white_to_move();
  BitBoard K,k,B,P;
  if(cp.black_pawn()){
    K=cp.black_king();
    B=cp.black_bishop();
    P=cp.black_pawn();
    k=cp.white_king();
    K.flip();k.flip();P.flip();B.flip();wtm=!wtm;
  }else{
    K=cp.white_king();
    B=cp.white_bishop();
    P=cp.white_pawn();
    k=cp.black_king();
  }
  if(P & kings_half){
    K.flop();k.flop();B.flop();P.flop();
  }
  if(P&a_file && distance_between(k.get_first_square(),a8)<2 && B&black_squares){r=0;}else{r=m_score;}
  
  return r;
}

int Evaluation::KRPKR()const
{
  int r{10};
  int rank,dtp;
  bool b_flip{false};
  BitBoard K,k,P,ps;
  if(cp.black_pawn()){
    K=cp.black_king();
    k=cp.white_king();
    P=cp.black_pawn();
    K.flip();k.flip();P.flip();b_flip=true;
  }else{
    K=cp.white_king();
    k=cp.black_king();
    P=cp.white_pawn();
  }

  rank=P.get_first_square().rank_number()+1;  // 1...8
  if(rank==2){rank=3;}
  dtp=8-rank;
  ps=P<<dtp*8;

  r+=(distance_between(k.get_unique_square(),ps)-dtp)*100;
  r-=distance_between(K.get_unique_square(),P)*10;
  if(r<=0){r=10;}
  
  if(b_flip){r=-r;}
  return r;
}

int Evaluation::KQKR()
{
  string s=cp.fen();
  int r=m_egtkqkr.eval(s);
  if(r!=0xFF){return r;}
  r=300;

  int sign{1};
  BitBoard losingking, winningking,losingrook;
  if(cp.black_queen()){
    sign=-1;
    losingking=cp.white_king();
    losingrook=cp.white_rook();
    winningking=cp.black_king();
  }else{
    losingking=cp.black_king();
    losingrook=cp.black_rook();
    winningking=cp.white_king();
  }

  if(losingking & border){r+=60;}
  else if(losingking & border_minus1){r+=40;}
  else if(losingking & border_minus2){r+=20;}

  r-=distance_between(losingking.get_unique_square(),winningking.get_unique_square());
  r+=distance_between(losingking.get_unique_square(),losingrook.get_unique_square());
  
  return r*sign;
}

//
//---------------------------------------------------------------------
