/*

Copyright (C) 2018 Werner Taelemans, Belgium

This file is part of Galjoen.

Galjoen is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; version 3 of the License.

Galjoen is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Galjoen; see the file COPYING.  If not, see
<http://www.gnu.org/licenses/>.

*/

#include<stdexcept>
#include<iostream>
#include<utility>
#include<cassert>
#include"../movegen/movegenerator.h"
#include"../search/analyze.h"
#include"../common/numerals.h"
using std::cout;
using std::endl;
using std::logic_error;
using std::swap;


MList::MList(bool x)
{
  m_allocate_mem=x;
  if(m_allocate_mem){
    m_vm=new Move[256];
  }
}

MList::~MList()
{
  if(m_allocate_mem){
    delete[] m_vm;
  }
}

void MList::set_vm(Move* pvm)
{
  m_vm=pvm;
}

Move* MList::get_vm()
{
  return m_vm;
}

Move& MList::operator[](int i)
{
  return m_vm[i];
}

void MList::set_chessposition(ChessPosition& cp)
{
  pcp=&cp;
}

void MList::set_analysis(Analysis* pa)
{
  pana=pa;
}

void MList::set_searchmoves(const vector<Move>& vsm)
{
  int i{0};
  Move vttm[256];
  for(size_t j=0;j<m_size;j++){
    for(auto sm : vsm){
      if(m_vm[j]==static_cast<Move>(sm)){vttm[i++]=m_vm[j];}
    }
  }
  m_size=i;
  for(size_t j=0;j<m_size;j++){m_vm[j]=vttm[j];}
}

void MList::set_hashtable_move(const Move& mh)
{
  m_htm=mh;
  m_htm.xcode(hashtable_move);
}

void MList::set_killers(const Move& ka,const Move& kb)
{
  m_ka=ka;
  m_kb=kb;
}

void MList::set_killers_2(const Move& ka,const Move& kb)
{
  m_ka_2=ka;
  m_kb_2=kb;
}

void MList::add(const Move& m)
{
  m_vm[m_size]=m;
  m_size++;
}

void MList::remove(const Move& m)
{
  for(size_t i=0;i<m_size;i++){
    if(m==m_vm[i]){
      for(size_t j=i;j<m_size-1;j++){
	m_vm[j]=m_vm[j+1];
      }
      m_size--;
      break;
    }
  }
}

void MList::set_log(bool x)
{
  m_log=x;
}

void MList::set_xcodes()
{
  int hm_v1{0},hm_v2{0},cm_v1{0},cm_v2{0},fm_v1{0},fm_v2{0};
  Move hm1,hm2,cm,cm1,cm2,fm1,fm2;
  
  // find best counter_move:
  if(pcp->last_move()!=NullMove()){
    const Move& m1=pcp->last_move();
    for(size_t i=0;i<m_size;i++){
      Move m2=m_vm[i];
      int cmv=pana->countermove().value(m1,m2);
      if(m2.xcode()==xnormal_move){
	if(cmv>cm_v1){
	  cm_v2=cm_v1;
	  cm_v1=cmv;
	  cm2=cm1;
	  cm1=m_vm[i];
	}else if(cmv>cm_v2){
	  cm_v2=cmv;
	  cm2=m2;
	}
      }
    }
  }
  
  // find best follow-up move:
  size_t hs{pcp->history().size()};
  if(hs>1 && pcp->history()[hs-2].last_move!=NullMove()){
    const Move& m1=pcp->history()[hs-2].last_move;

    for(size_t i=0;i<m_size;i++){
      Move m2=m_vm[i];
      if(m_vm[i].xcode()==xnormal_move){
	if(pana->countermove().value(m1,m2)>fm_v1){
	  fm_v2=fm_v1;
	  fm_v1=pana->countermove().value(m1,m2);
	  fm2=fm1;
	  fm1=m_vm[i];
	}else if(pana->countermove().value(m1,m2)>fm_v2){
	  fm_v2=pana->countermove().value(m1,m2);
	  fm2=m2;
	}
      }
    }
  }
  
  // find best historical moves:
  for(size_t i=0;i<m_size;i++){
    Move m2=m_vm[i];
    if(m2.xcode()==xnormal_move){
      if(pana->historytable().value(m2)>hm_v1){
	hm_v2=hm_v1;
	hm_v1=pana->historytable().value(m2);
	hm2=hm1;
	hm1=m2;
      }else if(pana->historytable().value(m2)>hm_v2){
	hm_v2=pana->historytable().value(m2);
	hm2=m2;
      }
    }
  }

  // set xcodes:
  for(size_t i=0;i<m_size;i++){
    if(m_vm[i]==m_htm){
      m_vm[i].xcode(hashtable_move);
    }else if(m_vm[i].capture()){
      if(m_vm[i].capture().abs_value()>=m_vm[i].piece().abs_value()){
	m_vm[i].xcode(good_capture);
      }else{
	int v=pana->see(m_vm[i]);
	if(v>=0){
	  m_vm[i].xcode(good_capture);
	}else{
	  m_vm[i].xcode(losing_capture);
	}
      }
    }else if(m_vm[i]==m_ka || m_vm[i]==m_kb){
      m_vm[i].xcode(killer);
    }else if(m_vm[i]==m_ka_2 || m_vm[i]==m_kb_2){
      m_vm[i].xcode(killer_2);
    }else if(m_vm[i]==cm1 || m_vm[i]==cm2){
      m_vm[i].xcode(counter_move);
    }else if(m_vm[i]==fm1 || m_vm[i]==fm2){
      m_vm[i].xcode(counter_move);
    }else if(m_vm[i]==hm1 || m_vm[i]==hm2){
      m_vm[i].xcode(historical_move);
    }else if(m_vm[i].to().bitboard()& centre){
      m_vm[i].xcode(to_centre);
    }else if(m_vm[i].to().bitboard()& bigcentre){
      m_vm[i].xcode(to_bigcentre);	
    }    
  }
  
  return;
}

void MList::sort()
{
  set_xcodes();
  
  // sort:
  for(unsigned int i=0;i<m_size;i++){
    switch(m_vm[i].xcode()){
    case hashtable_move:m_hashtable_move.push_back(m_vm[i]);break;
    case good_capture:m_good_capture.push_back(m_vm[i]);break;
    case killer:m_killer.push_back(m_vm[i]);break;
    case killer_2:m_killer_2.push_back(m_vm[i]);break;
    case counter_move:m_counter_move.push_back(m_vm[i]);break;
    case historical_move:m_historical_move.push_back(m_vm[i]);break;
    case fertility1:m_fertile_move_1.push_back(m_vm[i]);break;
    case fertility2:m_fertile_move_2.push_back(m_vm[i]);break;
    case fertility3:m_fertile_move_3.push_back(m_vm[i]);break;
    case to_centre:m_to_centre.push_back(m_vm[i]);break;
    case to_bigcentre:m_to_bigcentre.push_back(m_vm[i]);break;
    case losing_capture:m_losing_capture.push_back(m_vm[i]);break;
    case xnormal_move:m_xnormal_move.push_back(m_vm[i]);break;
    }
  }

  m_size=0;
  for(auto m : m_hashtable_move){m_vm[m_size++]=m;}
  for(auto m : m_good_capture){m_vm[m_size++]=m;}
  for(auto m : m_killer){m_vm[m_size++]=m;}
  for(auto m : m_counter_move){m_vm[m_size++]=m;}
  for(auto m : m_killer_2){m_vm[m_size++]=m;}
  for(auto m : m_historical_move){m_vm[m_size++]=m;}
  for(auto m : m_fertile_move_1){m_vm[m_size++]=m;}
  for(auto m : m_fertile_move_2){m_vm[m_size++]=m;}
  for(auto m : m_fertile_move_3){m_vm[m_size++]=m;}
  for(auto m : m_to_centre){m_vm[m_size++]=m;}
  for(auto m : m_to_bigcentre){m_vm[m_size++]=m;}
  for(auto m : m_losing_capture){m_vm[m_size++]=m;}
  for(auto m : m_xnormal_move){m_vm[m_size++]=m;}

  m_hashtable_move.clear();
  m_good_capture.clear();
  m_killer.clear();
  m_killer_2.clear();
  m_counter_move.clear();
  m_historical_move.clear();
  m_fertile_move_1.clear();
  m_fertile_move_2.clear();
  m_fertile_move_3.clear();
  m_to_centre.clear();
  m_to_bigcentre.clear();
  m_losing_capture.clear();
  m_xnormal_move.clear();

  return;
}

void MList::clear()
{
  m_htm=m_ka=m_kb=m_ka_2=m_kb_2=NullMove();
  m_size=0;
}

const Move& MList::first()
{
  m_current=0;
  // select best move:
  unsigned best=m_current;
  for(unsigned i=m_current+1;i<m_size;i++){
    if(m_vm[i].xcode()<m_vm[best].xcode()){
      best=i;
    }
  }

  // choose best move:
  swap(m_vm[0],m_vm[best]);
  return m_vm[0];
}

const Move& MList::next()
{
  m_current++;

  // select best move:
  unsigned best=m_current;
  for(unsigned i=m_current+1;i<m_size;i++){
    if(m_vm[i].xcode()<m_vm[best].xcode()){
      best=i;
    }
  }

  // choose best move:
  swap(m_vm[m_current],m_vm[best]);
  return m_vm[m_current];
}

Move& MList::back()
{
  return m_vm[m_size-1];
}

void MList::pop_back()
{
  m_size--;
}

void MList::push_back(const Move& x)
{
  m_vm[m_size++]=x;
}

void MList::insert(vector<Move>::iterator it_begin,vector<Move>::iterator it_end)
{
  for(auto i=it_begin;i!=it_end;i++){
    m_vm[m_size++]=*i;
  }
}

unsigned int MList::size()const
{
  return m_size;
}

string MList::dump()const
{
  ostringstream oss;
  oss << "{begin MoveList}" << endl;
  for(size_t i=0;i<m_size;i++){
    oss << "  " << m_vm[i].dump();    
  }
  oss << "{end MoveList}" << endl;
  return oss.str();
}
