/*

Copyright (C) 2018 Werner Taelemans, Belgium

This file is part of Galjoen.

Galjoen is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; version 3 of the License.

Galjoen is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Galjoen; see the file COPYING.  If not, see
<http://www.gnu.org/licenses/>.

*/

#ifndef ANALYZE_H
#define ANALYZE_H

#include<fstream>
#include<vector>
#include<utility>
#include<atomic>
#include<mutex>
#include"../movegen/mlist.h"
#include"killer.h"
#include"principal.h"
#include"multipv.h"
#include"../eval/evaluation.h"
#include"../board/hasher.h"
#include"history.h"
#include"countermove.h"
#include"lmr.h"
#include"../board/rep3filter.h"
#include"timemanager.h"
#include"../hashtable/transposition.h"
#include"../search/threadpool.h"
using std::vector;
using std::pair;
using std::atomic;
using std::mutex;
using std::ofstream;


class ChessPosition;


enum action_t {no_action,start_iterations,start_nws,stop_thinking};
enum status_t {idle,spinning,thinking,finished};


class Analysis
{  
 public:
  Analysis(ChessPosition&,TranspositionTable&, ThreadPool&);
  ~Analysis();
  Analysis& operator=(const Analysis&);
  void init(const ChessPosition&);

  int root_search(float,int,int);
  int pv_search(const float&,int,const int&,PrincipalVariation&);
  int nullmove_search(float,int,int);
  void iterative_search();
  int nullwindow_search(const float&,int,const int&,const int&,const int&);

  int see(Square);
  int see(const Move&);
  int quiescence(int,int,int=0);
  int static_eval();
  void iidsort(MoveList&,int,int,int);
  const CounterMove& countermove()const;
  const HistoryTable& historytable()const;
  Evaluation& evaluation();

  void clear_eval();
  void clear_hash();
  void hashtable_enabled(bool);
  void clear_killer();
  void clear_history();
  void clear_countermove();
  void set_log(bool);
  int eval()const;
  Move best_move()const;
  Move pondermove()const;
  void make_move(const Move&);
  void unmake_move(const Move&);
  void make_quies_move(const Move&);
  void unmake_quies_move(const Move&);
  void make_nullmove();
  void unmake_nullmove();
  bool nullmove_allowed()const;
  bool is_legal(TTmove&)const;
  void operator()();
  void set_multipv(unsigned int);
  void set_max_nodes(long long);
  void reset_max_nodes();
  void set_action(action_t);
  int get_status()const;
  void set_status(int);
  long long get_node_count()const;
  void clear_node_count();
  int get_time()const;
  int get_running_time()const;
  float get_ebf()const;
  int get_hashfull()const;
  int get_iteration()const;
  int get_seldepth()const;
  void set_aspiration(bool);
  
  void set_thread_alpha(int);
  void set_thread_beta(int);
  void set_thread_depth(float);
  void set_thread_lmr(int);
  void set_thread_reductions(int);
  void set_thread_extensions(int);
  const int& get_thread_score()const;
  const long long& get_thread_nodes()const;
  Duration get_thread_time();
  const int& get_thread_id()const;
  void set_thread_id(int);
  
  void quit();
  const ChessPosition& chessposition()const;
  PrincipalVariation pv()const;
  MultiPV mpv()const;
  string static_eval_report();
  string dump_line();

  //----------------- UCI commands: -----------------------
  //
  void reset();
  void set_wtime(int);
  void set_btime(int);
  void set_winc(int);
  void set_binc(int);
  void set_movestogo(int);
  void set_depth(int);
  void set_nodes(long long);
  void set_mate(int);
  void set_movetime(int);
  void set_infinite(bool);
  void set_pondering(bool);
  void set_searchmoves(vector<Move>);
  void set_hash_size(int);
  void set_uci_limitstrength(bool);
  void set_uci_elo(int);
  void ponderhit();
  //
  //-------------------------------------------------------


 private:
  // other objects:
  ChessPosition& cp;
  TranspositionTable& m_tt;
  ThreadPool& m_tp;
  Egt4Pieces m_egtkbnk;
  Egt4Pieces m_egtkqkr{white_queen_id,black_rook_id};
  Evaluation m_eva;
  PrincipalVariation m_pv;
  MultiPV m_mpv;
  XMoveList m_xml;
  Killer m_killer;
  HistoryTable m_ht;
  CounterMove m_cm;
  LmrTable m_lmr;
  TimeManager m_tman;
  Timer m_tm;

  // communication with engine:
  atomic<int> m_action;
  atomic<int> m_status;
  atomic<bool> m_quit;
  mutable mutex m_mutex_mpv;

  // SMP:
  int m_thread_score;
  float m_thread_depth;
  int m_thread_alpha;
  int m_thread_beta;
  int m_thread_lmr{0};
  int m_thread_reductions{0};
  int m_thread_extensions{0};
  long long m_thread_nodes{0};
  Timer m_thread_timer;
  int m_thread_id{-1};

  // coordinates:
  int m_iteration{0};
  int m_ply{0};
  int m_seldepth{0};
  int m_nullmove_level{0};
  unsigned char m_date{0};
  long long m_node_count{0};

  // statistics:
  uint64_t m_total_node_count{0};
  uint64_t m_total{0};
  uint64_t m_leave_nodes{0};
  uint64_t m_qnodes{0};
  uint64_t m_iid_sort{0};
  float m_ebf{0};
  int m_time;

  // logging:
  ofstream ofs_log;
  bool m_log{false};
  array<array<uint64_t,256>,max_search_depth> m_cut_off_frequency;         // cut_off_frequency[depth][move]
  array<array<uint64_t,256>,max_search_depth> m_cut_off_xmove_t;           // cut_off_xmove_type[depth][xmove_t]

  // max values:
  int m_max_depth;
  int m_sd_depth;
  long long m_max_nodes;

  // uci limits:
  vector<Move> m_searchmoves;
  bool m_uci_limitstrength{false};
  int m_uci_elo{2000};
  int m_elo_maxdepth{max_search_depth};
  int m_handicap{0};

  // result of analysis:
  int m_value;
  Move m_best_move;
  Move m_pondermove;

  // fertility and aspiration window:
  bool m_baspiration{true};
  array<pair<Move,uint64_t>,3> m_fertility;
  
  // private functions:
  void select_searchmoves(MoveList&);
  int limitstrength(int)const;
  void new_node();
  void put_in_tt(bound_t,int,int,Move,unsigned char);
  void set_fertility(const Move&,uint64_t);
  void sort_fertility(MoveList&);
};


#endif
